import time

import pytest
import requests

from garth import sso
from garth.auth_tokens import OAuth1Token, OAuth2Token
from garth.exc import GarthException, GarthHTTPError
from garth.http import Client


@pytest.mark.vcr
def test_login_email_password_fail(client: Client):
    with pytest.raises(GarthException):
        sso.login("user@example.com", "wrong_p@ssword", client=client)


@pytest.mark.vcr
def test_login_success(client: Client):
    oauth1, oauth2 = sso.login(
        "user@example.com", "correct_password", client=client
    )

    assert oauth1
    assert isinstance(oauth1, OAuth1Token)
    assert oauth2
    assert isinstance(oauth2, OAuth2Token)


@pytest.mark.vcr
def test_login_success_mfa(monkeypatch, client: Client):
    def mock_input(_):
        return "671091"

    monkeypatch.setattr("builtins.input", mock_input)
    oauth1, oauth2 = sso.login(
        "user@example.com", "correct_password", client=client
    )

    assert oauth1
    assert isinstance(oauth1, OAuth1Token)
    assert oauth2
    assert isinstance(oauth2, OAuth2Token)


@pytest.mark.vcr
def test_login_success_mfa_async(monkeypatch, client: Client):
    def mock_input(_):
        return "031174"

    async def prompt_mfa():
        return input("MFA code: ")

    monkeypatch.setattr("builtins.input", mock_input)
    oauth1, oauth2 = sso.login(
        "user@example.com",
        "correct_password",
        client=client,
        prompt_mfa=prompt_mfa,
    )

    assert oauth1
    assert isinstance(oauth1, OAuth1Token)
    assert oauth2
    assert isinstance(oauth2, OAuth2Token)


@pytest.mark.vcr
def test_login_mfa_fail(client: Client):
    with pytest.raises(GarthException):
        oauth1, oauth2 = sso.login(
            "user@example.com",
            "correct_password",
            client=client,
            prompt_mfa=lambda: "123456",
        )


@pytest.mark.vcr
def test_login_return_on_mfa(client: Client):
    result = sso.login(
        "user@example.com",
        "correct_password",
        client=client,
        return_on_mfa=True,
    )

    assert isinstance(result, tuple)
    result_type, client_state = result

    assert isinstance(client_state, dict)
    assert result_type == "needs_mfa"
    assert "login_params" in client_state
    assert "client" in client_state

    code = "123456"  # obtain from custom login

    # test resuming the login
    oauth1, oauth2 = sso.resume_login(client_state, code)

    assert oauth1
    assert isinstance(oauth1, OAuth1Token)
    assert oauth2
    assert isinstance(oauth2, OAuth2Token)


def test_set_expirations(oauth2_token_dict: dict):
    token = sso.set_expirations(oauth2_token_dict)
    assert (
        token["expires_at"] - time.time() - oauth2_token_dict["expires_in"] < 1
    )
    assert (
        token["refresh_token_expires_at"]
        - time.time()
        - oauth2_token_dict["refresh_token_expires_in"]
        < 1
    )


@pytest.mark.vcr
def test_exchange(authed_client: Client):
    assert authed_client.oauth1_token and isinstance(
        authed_client.oauth1_token, OAuth1Token
    )
    oauth1_token = authed_client.oauth1_token
    oauth2_token = sso.exchange(oauth1_token, client=authed_client)
    assert not oauth2_token.expired
    assert not oauth2_token.refresh_expired
    assert oauth2_token.token_type.title() == "Bearer"
    assert authed_client.oauth2_token != oauth2_token


def test_parse_sso_response_success():
    resp = {
        "serviceTicketId": "ST-123",
        "responseStatus": {
            "type": "SUCCESSFUL",
            "message": "",
            "httpStatus": "OK",
        },
    }
    result = sso._parse_sso_response(resp, "SUCCESSFUL")
    assert result["serviceTicketId"] == "ST-123"


def test_parse_sso_response_error():
    resp = {
        "responseStatus": {
            "type": "INVALID_CREDENTIALS",
            "message": "Bad password",
            "httpStatus": "UNAUTHORIZED",
        },
    }
    with pytest.raises(GarthException, match="INVALID_CREDENTIALS"):
        sso._parse_sso_response(resp, "SUCCESSFUL")


def test_parse_sso_response_no_expected():
    resp = {
        "responseStatus": {
            "type": "MFA_REQUIRED",
            "message": "",
            "httpStatus": "OK",
        },
    }
    result = sso._parse_sso_response(resp)
    assert result["responseStatus"]["type"] == "MFA_REQUIRED"


def test_handle_mfa_http_error(monkeypatch, client: Client):
    from unittest.mock import MagicMock

    def mock_post(*a, **kw):
        raise GarthHTTPError(
            msg="MFA verification failed",
            error=requests.HTTPError(response=MagicMock()),
        )

    monkeypatch.setattr(client, "post", mock_post)

    with pytest.raises(GarthHTTPError, match="MFA verification failed"):
        sso.handle_mfa(client, {"clientId": "X"}, lambda: "123456")


def test_login_unexpected_response_type(monkeypatch, client: Client):
    from unittest.mock import MagicMock

    unexpected_json = {
        "responseStatus": {
            "type": "CAPTCHA_REQUIRED",
            "message": "solve captcha",
            "httpStatus": "OK",
        },
    }

    call_count = 0

    def mock_request(method, *args, **kwargs):
        nonlocal call_count
        call_count += 1
        mock_resp = MagicMock()
        mock_resp.ok = True
        mock_resp.status_code = 200
        # First call is the cookie GET, second is the login POST
        if call_count == 2:
            mock_resp.json.return_value = unexpected_json
        client.last_resp = mock_resp
        return mock_resp

    monkeypatch.setattr(client, "request", mock_request)

    with pytest.raises(GarthException, match="CAPTCHA_REQUIRED"):
        sso.login("user@example.com", "password", client=client)



import json
import re
import uuid
from collections.abc import Callable
from typing import Any

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from requests import Response, Session

from .version import __version__


try:
    import logfire

    LOGFIRE_AVAILABLE = True
except ImportError:  # pragma: no cover
    logfire = None  # type: ignore[assignment]
    LOGFIRE_AVAILABLE = False


REDACTED = "[REDACTED]"
DEFAULT_TOKEN = "pylf_v1_us_6zTrqKw3YjJBHfpVPVr55wYcJ9sTV6sd4D3C4hsXLfJC"

_SENSITIVE_QUERY_PARAMS = [
    "username",
    "password",
    "refresh_token",
    "oauth_token",
    "oauth_token_secret",
    "mfa_token",
    "embed",
]

# Pre-compiled regex patterns for query params (key=value format)
_QUERY_PARAM_PATTERNS = [
    re.compile(rf"{key}=[^&\s]*", re.IGNORECASE)
    for key in _SENSITIVE_QUERY_PARAMS
]
_QUERY_PARAM_REPLACEMENTS = [
    f"{key}={REDACTED}" for key in _SENSITIVE_QUERY_PARAMS
]

# JSON field names to sanitize
JSON_SENSITIVE_FIELDS = [
    "access_token",
    "refresh_token",
    "jti",
    "consumer_key",
    "consumer_secret",
    "password",
    "oauth_token",
    "oauth_token_secret",
]

# Header keys to sanitize
SENSITIVE_HEADERS = ["Authorization", "Cookie", "Set-Cookie"]
SENSITIVE_HEADERS_LOWER = {h.lower() for h in SENSITIVE_HEADERS}

# Pre-compiled regex for cookie sanitization
_COOKIE_PATTERN = re.compile(r"=[^;]*")


def sanitize_cookie(cookie_value: str) -> str:
    """Sanitize cookie value by redacting all values."""
    return _COOKIE_PATTERN.sub(f"={REDACTED}", cookie_value)


def sanitize(text: str) -> str:
    """Sanitize sensitive data from request/response text."""
    # Sanitize query params using pre-compiled patterns
    for pattern, replacement in zip(
        _QUERY_PARAM_PATTERNS, _QUERY_PARAM_REPLACEMENTS, strict=True
    ):
        text = pattern.sub(replacement, text)

    # Try to sanitize JSON
    try:
        data = json.loads(text)
        if isinstance(data, dict):
            for fld in JSON_SENSITIVE_FIELDS:
                if fld in data:
                    data[fld] = REDACTED
            return json.dumps(data)
    except (json.JSONDecodeError, TypeError):
        pass

    return text


def sanitize_headers(headers: dict) -> dict:
    """Sanitize sensitive headers."""
    sanitized = dict(headers)
    for key in list(sanitized.keys()):
        if key.lower() in SENSITIVE_HEADERS_LOWER:
            sanitized[key] = REDACTED
    return sanitized


def _scrubbing_callback(m):
    """Bypass logfire scrubbing since we've already sanitized all data."""
    return m.value


class Telemetry(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="GARTH_TELEMETRY_",
        extra="ignore",
    )

    enabled: bool = True
    send_to_logfire: bool = True
    token: str = DEFAULT_TOKEN
    callback: Callable[[dict], None] | None = Field(default=None, exclude=True)
    session_id: str = Field(
        default_factory=lambda: uuid.uuid4().hex[:16],
        exclude=True,
    )
    _logfire_configured: bool = False
    _logfire_instance: Any = None
    _attached_sessions: set = set()

    def model_post_init(self, __context):
        self._attached_sessions = set()

    def _default_callback(self, data: dict):
        """Default callback that sends to logfire."""
        if not LOGFIRE_AVAILABLE or not self._logfire_configured:
            return
        self._logfire_instance.info(
            "http {method} {url} {status_code}", **data
        )

    def _response_hook(self, response: Response, *args, **kwargs):
        """Session hook that captures request/response data."""
        if not self.enabled:
            return

        try:
            request = response.request
            if request is None:  # pragma: no cover
                return
            data = {
                "session_id": self.session_id,
                "garth_version": __version__,
                "method": request.method,
                "url": request.url,
                "status_code": response.status_code,
                "request_headers": str(
                    sanitize_headers(dict(request.headers))
                ),
                "response_headers": str(
                    sanitize_headers(dict(response.headers))
                ),
            }

            if request.body:
                body = request.body
                if isinstance(body, bytes):
                    body = body.decode("utf-8", errors="replace")
                data["request_body"] = sanitize(body)

            data["response_body"] = sanitize(response.text)

            callback = self.callback or self._default_callback
            callback(data)
        except Exception:
            pass  # Don't let telemetry errors break the app

    def configure(
        self,
        enabled: bool | None = None,
        send_to_logfire: bool | None = None,
        token: str | None = None,
        callback: Callable[[dict], None] | None = None,
    ):
        """
        Configure telemetry. Enabled by default.

        Args:
            enabled: Enable/disable telemetry (default: True)
            send_to_logfire: Send to Logfire Cloud (default: True)
            token: Logfire write token
            callback: Custom callback for telemetry data. If provided,
                logfire will not be configured and data will be passed
                to this callback instead.
        """
        if enabled is not None:
            self.enabled = enabled
        if send_to_logfire is not None:
            self.send_to_logfire = send_to_logfire
        if token is not None:
            self.token = token
        if callback is not None:
            self.callback = callback

        if not self.enabled:
            return

        # Configure logfire only if using default callback and not yet done
        if self.callback is None and not self._logfire_configured:
            self._configure_logfire()

    def _configure_logfire(self):
        """Configure a local logfire instance for garth telemetry.

        Uses local=True to avoid overwriting logfire configuration
        in applications that integrate garth.
        """
        if not LOGFIRE_AVAILABLE:
            return

        self._logfire_instance = logfire.configure(
            local=True,
            service_name="garth",
            send_to_logfire=self.send_to_logfire,
            token=self.token,
            scrubbing=logfire.ScrubbingOptions(callback=_scrubbing_callback),
            console=False,
        )
        self._logfire_configured = True

    def attach(self, session: Session):
        """Attach telemetry hooks to a session.

        This method is idempotent - calling it multiple times with the
        same session will only attach the hook once.
        """
        session_id = id(session)
        if session_id in self._attached_sessions:
            return

        session.hooks["response"].append(self._response_hook)
        self._attached_sessions.add(session_id)



from __future__ import annotations

import builtins
from datetime import date, datetime

from pydantic.dataclasses import dataclass
from typing_extensions import Self

from .. import http
from ..utils import camel_to_snake_dict, format_end_date
from ._base import Data


@dataclass
class Baseline:
    low_upper: int
    balanced_low: int
    balanced_upper: int
    marker_value: float


@dataclass
class HRVSummary:
    calendar_date: date
    weekly_avg: int
    baseline: Baseline
    status: str
    feedback_phrase: str
    create_time_stamp: datetime
    last_night_avg: int | None = None
    last_night_5_min_high: int | None = None


@dataclass
class HRVReading:
    hrv_value: int
    reading_time_gmt: datetime
    reading_time_local: datetime


@dataclass
class HRVData(Data):
    user_profile_pk: int
    hrv_summary: HRVSummary
    hrv_readings: builtins.list[HRVReading]
    start_timestamp_gmt: datetime
    end_timestamp_gmt: datetime
    start_timestamp_local: datetime
    end_timestamp_local: datetime
    sleep_start_timestamp_gmt: datetime | None = None
    sleep_end_timestamp_gmt: datetime | None = None
    sleep_start_timestamp_local: datetime | None = None
    sleep_end_timestamp_local: datetime | None = None

    @classmethod
    def get(
        cls,
        day: date | str | None = None,
        *,
        client: http.Client | None = None,
    ) -> Self | None:
        client = client or http.client
        day = format_end_date(day)
        path = f"/hrv-service/hrv/{day}"
        hrv_data = client.connectapi(path)
        if not hrv_data:
            return None
        assert isinstance(hrv_data, dict), (
            f"Expected dict from {path}, got {type(hrv_data).__name__}"
        )
        hrv_data = camel_to_snake_dict(hrv_data)
        return cls(**hrv_data)

    @classmethod
    def list(cls, *args, **kwargs) -> builtins.list[Self]:
        data = super().list(*args, **kwargs)
        return sorted(data, key=lambda d: d.hrv_summary.calendar_date)



from __future__ import annotations

import asyncio
import inspect
import time
from collections.abc import Callable
from typing import Any, Literal
from urllib.parse import parse_qs

import requests
from requests import Session
from requests_oauthlib import OAuth1Session

from . import http
from .auth_tokens import OAuth1Token, OAuth2Token
from .exc import GarthException


CLIENT_ID = "GCM_ANDROID_DARK"
OAUTH_CONSUMER_URL = "https://thegarth.s3.amazonaws.com/oauth_consumer.json"
OAUTH_CONSUMER: dict[str, str] = {}

SSO_SUCCESSFUL = "SUCCESSFUL"
SSO_MFA_REQUIRED = "MFA_REQUIRED"

# Android user agent — must match the Android consumer key from S3
OAUTH_USER_AGENT = {"User-Agent": "com.garmin.android.apps.connectmobile"}

# Browser-like headers for SSO to avoid Cloudflare challenges.
# The SSO endpoints run in a WebView, so requests must look like a
# browser — not a Python HTTP client.
_SSO_UA = (
    "Mozilla/5.0 (iPhone; CPU iPhone OS 18_7 like Mac OS X) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"
)
SSO_PAGE_HEADERS = {
    "User-Agent": _SSO_UA,
    "Accept": (
        "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
    ),
    "Accept-Language": "en-US,en;q=0.9",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Dest": "document",
}


class GarminOAuth1Session(OAuth1Session):
    def __init__(
        self,
        /,
        parent: Session | None = None,
        **kwargs,
    ):
        global OAUTH_CONSUMER
        if not OAUTH_CONSUMER:
            request_kwargs: dict[str, Any] = {}
            if parent is not None:
                request_kwargs["proxies"] = parent.proxies
                request_kwargs["verify"] = parent.verify
            OAUTH_CONSUMER = requests.get(
                OAUTH_CONSUMER_URL, **request_kwargs
            ).json()
        super().__init__(
            OAUTH_CONSUMER["consumer_key"],
            OAUTH_CONSUMER["consumer_secret"],
            **kwargs,
        )
        if parent is not None:
            self.mount("https://", parent.adapters["https://"])
            self.cookies.update(parent.cookies)
            self.proxies = parent.proxies
            self.verify = parent.verify
            self.hooks["response"].extend(parent.hooks["response"])


def _parse_sso_response(
    resp_json: dict[str, Any],
    expected_type: str | None = None,
) -> dict[str, Any]:
    status = resp_json.get("responseStatus", {})
    resp_type = status.get("type", "UNKNOWN")
    message = status.get("message", "")
    if expected_type and resp_type != expected_type:
        detail = f"{resp_type}: {message}" if message else resp_type
        raise GarthException(msg=f"SSO error: {detail}")
    return resp_json


def login(
    email: str,
    password: str,
    /,
    client: http.Client | None = None,
    prompt_mfa: Callable | None = lambda: input("MFA code: "),
    return_on_mfa: bool = False,
) -> (
    tuple[OAuth1Token, OAuth2Token]
    | tuple[Literal["needs_mfa"], dict[str, Any]]
):
    client = client or http.client
    service_url = f"https://mobile.integration.{client.domain}/gcm/android"
    login_params = {
        "clientId": CLIENT_ID,
        "locale": "en-US",
        "service": service_url,
    }

    # Set cookies
    client.get(
        "sso",
        "/mobile/sso/en/sign-in",
        params={"clientId": CLIENT_ID},
        headers={**SSO_PAGE_HEADERS, "Sec-Fetch-Site": "none"},
    )

    # Submit login
    client.post(
        "sso",
        "/mobile/api/login",
        params=login_params,
        headers=SSO_PAGE_HEADERS,
        json={
            "username": email,
            "password": password,
            "rememberMe": False,
            "captchaToken": "",
        },
    )
    resp_json = client.last_resp.json()
    resp_type = resp_json.get("responseStatus", {}).get("type")

    if resp_type == SSO_SUCCESSFUL:
        ticket = resp_json["serviceTicketId"]
        return _complete_login(ticket, client)

    if resp_type == SSO_MFA_REQUIRED:
        mfa_info = resp_json.get("customerMfaInfo") or {}
        mfa_method = mfa_info.get("mfaLastMethodUsed") or "email"
        if return_on_mfa or prompt_mfa is None:
            return "needs_mfa", {
                "login_params": login_params,
                "client": client,
                "mfa_method": mfa_method,
            }
        ticket = handle_mfa(
            client, login_params, prompt_mfa, mfa_method=mfa_method
        )
        return _complete_login(ticket, client)

    _parse_sso_response(resp_json, SSO_SUCCESSFUL)
    raise GarthException(msg="Unexpected SSO response")  # pragma: no cover


def get_oauth1_token(ticket: str, client: http.Client) -> OAuth1Token:
    sess = GarminOAuth1Session(parent=client.sess)
    base_url = f"https://connectapi.{client.domain}/oauth-service/oauth/"
    login_url = f"https://mobile.integration.{client.domain}/gcm/android"
    url = (
        f"{base_url}preauthorized?ticket={ticket}&login-url={login_url}"
        "&accepts-mfa-tokens=true"
    )
    resp = sess.get(
        url,
        headers=OAUTH_USER_AGENT,
        timeout=client.timeout,
    )
    resp.raise_for_status()
    parsed = parse_qs(resp.text)
    token = {k: v[0] for k, v in parsed.items()}
    return OAuth1Token(domain=client.domain, **token)  # type: ignore


def exchange(
    oauth1: OAuth1Token,
    client: http.Client,
    *,
    login: bool = False,
) -> OAuth2Token:
    sess = GarminOAuth1Session(
        resource_owner_key=oauth1.oauth_token,
        resource_owner_secret=oauth1.oauth_token_secret,
        parent=client.sess,
    )
    data: dict[str, str] = {}
    if login:
        data["audience"] = "GARMIN_CONNECT_MOBILE_ANDROID_DI"
    if oauth1.mfa_token:
        data["mfa_token"] = oauth1.mfa_token
    base_url = f"https://connectapi.{client.domain}/oauth-service/oauth/"
    url = f"{base_url}exchange/user/2.0"
    headers = {
        **OAUTH_USER_AGENT,
        **{"Content-Type": "application/x-www-form-urlencoded"},
    }
    resp = sess.post(
        url,
        headers=headers,
        data=data,
        timeout=client.timeout,
    )
    resp.raise_for_status()
    token = resp.json()
    return OAuth2Token(**set_expirations(token))


def handle_mfa(
    client: http.Client,
    login_params: dict,
    prompt_mfa: Callable,
    *,
    mfa_method: str = "email",
) -> str:
    if inspect.iscoroutinefunction(prompt_mfa):
        mfa_code = asyncio.run(prompt_mfa())
    else:
        mfa_code = prompt_mfa()
    client.post(
        "sso",
        "/mobile/api/mfa/verifyCode",
        params=login_params,
        headers=SSO_PAGE_HEADERS,
        json={
            "mfaMethod": mfa_method,
            "mfaVerificationCode": mfa_code,
            "rememberMyBrowser": False,
            "reconsentList": [],
            "mfaSetup": False,
        },
    )
    resp_json = _parse_sso_response(client.last_resp.json(), SSO_SUCCESSFUL)
    return resp_json["serviceTicketId"]


def set_expirations(token: dict) -> dict:
    token["expires_at"] = int(time.time() + token["expires_in"])
    token["refresh_token_expires_at"] = int(
        time.time() + token["refresh_token_expires_in"]
    )
    return token


def resume_login(
    client_state: dict, mfa_code: str
) -> tuple[OAuth1Token, OAuth2Token]:
    client = client_state["client"]
    login_params = client_state["login_params"]
    mfa_method = client_state.get("mfa_method", "email")
    ticket = handle_mfa(
        client, login_params, lambda: mfa_code, mfa_method=mfa_method
    )
    return _complete_login(ticket, client)


def _complete_login(
    ticket: str, client: http.Client
) -> tuple[OAuth1Token, OAuth2Token]:
    # Sets Cloudflare LB cookie for backend pinning — best-effort
    try:
        client.get(
            "sso",
            "/portal/sso/embed",
            headers={**SSO_PAGE_HEADERS, "Sec-Fetch-Site": "same-origin"},
            referrer=True,
        )
    except GarthException:  # pragma: no cover
        pass

    oauth1 = get_oauth1_token(ticket, client)
    oauth2 = exchange(oauth1, client, login=True)
    return oauth1, oauth2
